from functools import lru_cache

from pydantic import Field
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict


class OpenAPISchemaValidatorSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="OPENAPI_SCHEMA_VALIDATOR_",
        extra="ignore",
    )

    compiled_validator_cache_max_size: int = Field(default=128, ge=0)


@lru_cache(maxsize=1)
def get_settings() -> OpenAPISchemaValidatorSettings:
    return OpenAPISchemaValidatorSettings()


def reset_settings_cache() -> None:
    get_settings.cache_clear()



from typing import Any

from jsonschema.validators import validates

from openapi_schema_validator._specifications import (
    REGISTRY as OPENAPI_SPECIFICATIONS,
)

__all__ = [
    "OAS31_BASE_DIALECT_ID",
    "OAS31_BASE_DIALECT_METASCHEMA",
    "OAS32_BASE_DIALECT_ID",
    "OAS32_BASE_DIALECT_METASCHEMA",
    "register_openapi_dialect",
]

OAS31_BASE_DIALECT_ID = "https://spec.openapis.org/oas/3.1/dialect/base"
OAS31_BASE_DIALECT_METASCHEMA = OPENAPI_SPECIFICATIONS.contents(
    OAS31_BASE_DIALECT_ID,
)
OAS32_BASE_DIALECT_ID = "https://spec.openapis.org/oas/3.2/dialect/2025-09-17"
OAS32_BASE_DIALECT_METASCHEMA = OPENAPI_SPECIFICATIONS.contents(
    OAS32_BASE_DIALECT_ID,
)

_REGISTERED_VALIDATORS: dict[tuple[str, str], Any] = {}


def register_openapi_dialect(
    *,
    validator: Any,
    dialect_id: str,
    version_name: str,
    metaschema: Any,
) -> Any:
    key = (dialect_id, version_name)
    registered_validator = _REGISTERED_VALIDATORS.get(key)

    if registered_validator is validator:
        return validator
    if registered_validator is not None:
        return registered_validator

    validator.META_SCHEMA = metaschema
    validator = validates(version_name)(validator)
    _REGISTERED_VALIDATORS[key] = validator
    return validator



import re
import sys
from pathlib import Path

ROOT_DIR = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(ROOT_DIR))


def _read_project_version() -> str:
    pyproject_path = ROOT_DIR / "pyproject.toml"
    pyproject_content = pyproject_path.read_text(encoding="utf-8")
    match = re.search(
        r"\[tool\.poetry\][\s\S]*?^version\s*=\s*\"([^\"]+)\"",
        pyproject_content,
        re.MULTILINE,
    )
    if match is None:
        return "unknown"
    return match.group(1)


project = "openapi-schema-validator"
copyright = "2023, Artur Maciag"
author = "Artur Maciag"

release = _read_project_version()

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx_immaterial",
]

templates_path = ["_templates"]

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

html_theme = "sphinx_immaterial"

html_static_path = []

html_title = "openapi-schema-validator"

html_theme_options = {
    "analytics": {
        "provider": "google",
        "property": "G-11RDPBZ7EJ",
    },
    "repo_url": "https://github.com/python-openapi/openapi-schema-validator/",
    "repo_name": "openapi-schema-validator",
    "icon": {
        "repo": "fontawesome/brands/github-alt",
        "edit": "material/file-edit-outline",
    },
    "palette": [
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "lime",
            "accent": "amber",
            "scheme": "slate",
            "toggle": {
                "icon": "material/toggle-switch",
                "name": "Switch to light mode",
            },
        },
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "lime",
            "accent": "amber",
            "toggle": {
                "icon": "material/toggle-switch-off-outline",
                "name": "Switch to dark mode",
            },
        },
    ],
    "globaltoc_collapse": False,
}



import re
from typing import Any

_REGEX_CLASS: Any = None
_REGRESS_ERROR: type[Exception] = Exception

try:
    from regress import Regex as _REGEX_CLASS
    from regress import RegressError as _REGRESS_ERROR
except ImportError:  # pragma: no cover - optional dependency
    pass


class ECMARegexSyntaxError(ValueError):
    pass


def has_ecma_regex() -> bool:
    return _REGEX_CLASS is not None


def is_valid_regex(pattern: str) -> bool:
    if _REGEX_CLASS is None:
        try:
            re.compile(pattern)
        except re.error:
            return False
        return True

    try:
        _REGEX_CLASS(pattern)
    except _REGRESS_ERROR:
        return False
    return True


def search(pattern: str, instance: str) -> bool:
    if _REGEX_CLASS is None:
        return re.search(pattern, instance) is not None

    try:
        return _REGEX_CLASS(pattern).find(instance) is not None
    except _REGRESS_ERROR as exc:
        raise ECMARegexSyntaxError(str(exc)) from exc



from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import Any

LOWER_IS_BETTER_METRICS = {
    "compile_ms",
    "first_validate_ms",
    "compiled_peak_memory_kib",
}
HIGHER_IS_BETTER_METRICS = {
    "compiled_validations_per_second",
    "helper_validations_per_second",
    "helper_trusted_validations_per_second",
}
ALL_METRICS = [
    "compile_ms",
    "first_validate_ms",
    "compiled_validations_per_second",
    "helper_validations_per_second",
    "helper_trusted_validations_per_second",
    "compiled_peak_memory_kib",
]


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Compare two benchmark JSON reports.",
    )
    parser.add_argument(
        "--baseline",
        type=Path,
        required=True,
        help="Path to baseline benchmark JSON.",
    )
    parser.add_argument(
        "--candidate",
        type=Path,
        required=True,
        help="Path to candidate benchmark JSON.",
    )
    parser.add_argument(
        "--regression-threshold",
        type=float,
        default=0.0,
        help=(
            "Percent threshold for regressions. "
            "Example: 5 means fail only when regression exceeds 5%%."
        ),
    )
    parser.add_argument(
        "--fail-on-regression",
        action="store_true",
        help="Exit with status 1 if regressions exceed threshold.",
    )
    return parser.parse_args()


def _load_report(path: Path) -> dict[str, Any]:
    return json.loads(path.read_text(encoding="utf-8"))


def _cases_by_name(report: dict[str, Any]) -> dict[str, dict[str, Any]]:
    return {case["name"]: case for case in report["cases"]}


def _percent_change(baseline_value: float, candidate_value: float) -> float:
    if baseline_value == 0:
        if candidate_value == 0:
            return 0.0
        return float("inf")
    return ((candidate_value - baseline_value) / baseline_value) * 100.0


def _is_regression(metric: str, percent_change: float) -> bool:
    if metric in LOWER_IS_BETTER_METRICS:
        return percent_change > 0
    return percent_change < 0


def _format_status(is_regression: bool, percent_change: float) -> str:
    if abs(percent_change) < 1e-12:
        return "no change (0.00%)"

    direction = "regression" if is_regression else "improvement"
    sign = "+" if percent_change >= 0 else ""
    return f"{direction} ({sign}{percent_change:.2f}%)"


def _compare_reports(
    baseline: dict[str, Any],
    candidate: dict[str, Any],
    regression_threshold: float,
) -> tuple[list[str], list[str]]:
    baseline_cases = _cases_by_name(baseline)
    candidate_cases = _cases_by_name(candidate)

    report_lines: list[str] = []
    regressions: list[str] = []

    for case_name in sorted(baseline_cases):
        if case_name not in candidate_cases:
            regressions.append(
                f"Missing case in candidate report: {case_name}"
            )
            continue

        report_lines.append(f"Case: {case_name}")
        baseline_case = baseline_cases[case_name]
        candidate_case = candidate_cases[case_name]

        for metric in ALL_METRICS:
            baseline_value = float(baseline_case[metric])
            candidate_value = float(candidate_case[metric])
            change = _percent_change(baseline_value, candidate_value)
            regression = _is_regression(metric, change)
            status = _format_status(regression, change)

            report_lines.append(
                "  "
                f"{metric}: baseline={baseline_value:.6f} "
                f"candidate={candidate_value:.6f} -> {status}"
            )

            if regression and abs(change) > regression_threshold:
                regressions.append(
                    f"{case_name} {metric} regressed by {abs(change):.2f}%"
                )

    extra_candidate_cases = set(candidate_cases).difference(baseline_cases)
    for case_name in sorted(extra_candidate_cases):
        report_lines.append(f"Case present only in candidate: {case_name}")

    return report_lines, regressions


def main() -> int:
    args = _parse_args()
    baseline = _load_report(args.baseline)
    candidate = _load_report(args.candidate)
    report_lines, regressions = _compare_reports(
        baseline,
        candidate,
        args.regression_threshold,
    )

    print(
        f"Comparing candidate {args.candidate} "
        f"against baseline {args.baseline}"
    )
    print("")
    print("\n".join(report_lines))

    if regressions:
        print("")
        print("Regressions above threshold:")
        for regression in regressions:
            print(f"- {regression}")

        if args.fail_on_regression:
            return 1

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from typing import Any
from typing import Iterator
from typing import Mapping
from typing import cast

from jsonschema._keywords import allOf as _allOf
from jsonschema._keywords import anyOf as _anyOf
from jsonschema._keywords import oneOf as _oneOf
from jsonschema._keywords import pattern as _pattern
from jsonschema._utils import extras_msg
from jsonschema._utils import find_additional_properties
from jsonschema.exceptions import FormatError
from jsonschema.exceptions import ValidationError
from jsonschema.exceptions import _WrappedReferencingError

from openapi_schema_validator._regex import ECMARegexSyntaxError
from openapi_schema_validator._regex import has_ecma_regex
from openapi_schema_validator._regex import search as regex_search


def handle_discriminator(
    validator: Any, _: Any, instance: Any, schema: Mapping[str, Any]
) -> Iterator[ValidationError]:
    """
    Handle presence of discriminator in anyOf, oneOf and allOf.
    The behaviour is the same in all 3 cases because at most 1 schema will match.
    """
    discriminator = schema["discriminator"]
    prop_name = discriminator["propertyName"]

    if not validator.is_type(instance, "object"):
        yield ValidationError(
            f"{instance!r} is not of type 'object'", context=[]
        )
        return

    prop_value = instance.get(prop_name)
    if not prop_value:
        # instance is missing $propertyName
        yield ValidationError(
            f"{instance!r} does not contain discriminating property {prop_name!r}",
            context=[],
        )
        return

    # Use explicit mapping if available, otherwise try implicit value
    ref = (
        discriminator.get("mapping", {}).get(prop_value)
        or f"#/components/schemas/{prop_value}"
    )

    if not isinstance(ref, str):
        # this is a schema error
        yield ValidationError(
            f"{instance!r} mapped value for {prop_value!r} should be a string, was {ref!r}",
            context=[],
        )
        return

    try:
        validator._validate_reference(ref=ref, instance=instance)
    except _WrappedReferencingError:
        yield ValidationError(
            f"{instance!r} reference {ref!r} could not be resolved",
            context=[],
        )
        return

    yield from validator.descend(instance, {"$ref": ref})


def anyOf(
    validator: Any,
    anyOf: list[Mapping[str, Any]],
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if "discriminator" not in schema:
        yield from cast(
            Iterator[ValidationError],
            _anyOf(validator, anyOf, instance, schema),
        )
    else:
        yield from handle_discriminator(validator, anyOf, instance, schema)


def oneOf(
    validator: Any,
    oneOf: list[Mapping[str, Any]],
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if "discriminator" not in schema:
        yield from cast(
            Iterator[ValidationError],
            _oneOf(validator, oneOf, instance, schema),
        )
    else:
        yield from handle_discriminator(validator, oneOf, instance, schema)


def allOf(
    validator: Any,
    allOf: list[Mapping[str, Any]],
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if "discriminator" not in schema:
        yield from cast(
            Iterator[ValidationError],
            _allOf(validator, allOf, instance, schema),
        )
    else:
        yield from handle_discriminator(validator, allOf, instance, schema)


def type(
    validator: Any,
    data_type: str,
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    """Default type validator - allows Python bytes for binary format for pragmatic reasons."""
    if instance is None:
        # nullable implementation based on OAS 3.0.3
        # * nullable is only meaningful if its value is true
        # * nullable: true is only meaningful in combination with a type
        #   assertion specified in the same Schema Object.
        # * nullable: true operates within a single Schema Object
        if schema.get("nullable") is True:
            return
        yield ValidationError("None for not nullable")

    # Pragmatic: allow bytes for binary format (common in Python use cases)
    if (
        data_type == "string"
        and schema.get("format") == "binary"
        and isinstance(instance, bytes)
    ):
        return

    if not validator.is_type(instance, data_type):
        data_repr = repr(data_type)
        yield ValidationError(f"{instance!r} is not of type {data_repr}")


def strict_type(
    validator: Any,
    data_type: str,
    instance: Any,
    schema: Any,
) -> Any:
    """
    Strict type validator - follows OAS spec precisely.
    Does NOT allow Python bytes for binary format.
    """
    if instance is None:
        if schema.get("nullable") is True:
            return
        yield ValidationError("None for not nullable")

    if not validator.is_type(instance, data_type):
        data_repr = repr(data_type)
        yield ValidationError(f"{instance!r} is not of type {data_repr}")


def pattern(
    validator: Any,
    patrn: str,
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if not has_ecma_regex():
        yield from cast(
            Iterator[ValidationError],
            _pattern(validator, patrn, instance, schema),
        )
        return

    if not validator.is_type(instance, "string"):
        return

    try:
        matches = regex_search(patrn, instance)
    except ECMARegexSyntaxError as exc:
        yield ValidationError(
            f"{patrn!r} is not a valid regular expression ({exc})"
        )
        return

    if not matches:
        yield ValidationError(f"{instance!r} does not match {patrn!r}")


def format(
    validator: Any,
    format: str,
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if instance is None:
        return

    if validator.format_checker is not None:
        try:
            validator.format_checker.check(instance, format)
        except FormatError as error:
            yield ValidationError(str(error), cause=error.cause)


def items(
    validator: Any,
    items: Mapping[str, Any],
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if not validator.is_type(instance, "array"):
        return

    for index, item in enumerate(instance):
        yield from validator.descend(item, items, path=index)


def required(
    validator: Any,
    required: list[str],
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if not validator.is_type(instance, "object"):
        return
    for property in required:
        if property not in instance:
            prop_schema = schema.get("properties", {}).get(property)
            if prop_schema:
                read_only = prop_schema.get("readOnly", False)
                write_only = prop_schema.get("writeOnly", False)
                if (
                    getattr(validator, "write", True)
                    and read_only
                    or getattr(validator, "read", True)
                    and write_only
                ):
                    continue
            yield ValidationError(f"{property!r} is a required property")


def read_required(
    validator: Any,
    required: list[str],
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if not validator.is_type(instance, "object"):
        return
    for property in required:
        if property not in instance:
            prop_schema = schema.get("properties", {}).get(property)
            if prop_schema:
                write_only = prop_schema.get("writeOnly", False)
                if getattr(validator, "read", True) and write_only:
                    continue
            yield ValidationError(f"{property!r} is a required property")


def write_required(
    validator: Any,
    required: list[str],
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if not validator.is_type(instance, "object"):
        return
    for property in required:
        if property not in instance:
            prop_schema = schema.get("properties", {}).get(property)
            if prop_schema:
                read_only = prop_schema.get("readOnly", False)
                if read_only:
                    continue
            yield ValidationError(f"{property!r} is a required property")


def additionalProperties(
    validator: Any,
    aP: Any,
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if not validator.is_type(instance, "object"):
        return

    extras = set(find_additional_properties(instance, schema))

    if not extras:
        return

    if validator.is_type(aP, "object"):
        for extra in extras:
            for error in validator.descend(instance[extra], aP, path=extra):
                yield error
    elif validator.is_type(aP, "boolean"):
        if not aP:
            error = "Additional properties are not allowed (%s %s unexpected)"
            yield ValidationError(error % extras_msg(extras))


def write_readOnly(
    validator: Any,
    ro: bool,
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if not ro:
        return
    yield ValidationError(f"Tried to write read-only property with {instance}")


def read_writeOnly(
    validator: Any,
    wo: bool,
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    if not wo:
        return
    yield ValidationError(f"Tried to read write-only property with {instance}")


def not_implemented(
    validator: Any,
    value: Any,
    instance: Any,
    schema: Mapping[str, Any],
) -> Iterator[ValidationError]:
    yield from ()



from typing import Any
from typing import cast

from jsonschema._types import TypeChecker
from jsonschema._types import draft202012_type_checker
from jsonschema._types import is_array
from jsonschema._types import is_bool
from jsonschema._types import is_integer
from jsonschema._types import is_number
from jsonschema._types import is_object


def is_string(checker: Any, instance: Any) -> bool:
    # Both strict and pragmatic: only accepts str for plain string type
    return isinstance(instance, str)


oas30_type_checker = TypeChecker(
    cast(
        Any,
        {
            "string": is_string,
            "number": is_number,
            "integer": is_integer,
            "boolean": is_bool,
            "array": is_array,
            "object": is_object,
        },
    ),
)

oas31_type_checker = draft202012_type_checker



from openapi_schema_validator.settings import get_settings
from openapi_schema_validator.settings import reset_settings_cache


def test_compiled_validator_cache_size_env_is_cached(monkeypatch):
    monkeypatch.setenv(
        "OPENAPI_SCHEMA_VALIDATOR_COMPILED_VALIDATOR_CACHE_MAX_SIZE",
        "11",
    )
    reset_settings_cache()

    first = get_settings()
    assert first.compiled_validator_cache_max_size == 11

    monkeypatch.setenv(
        "OPENAPI_SCHEMA_VALIDATOR_COMPILED_VALIDATOR_CACHE_MAX_SIZE",
        "3",
    )
    second = get_settings()
    assert second.compiled_validator_cache_max_size == 11

    reset_settings_cache()
    third = get_settings()
    assert third.compiled_validator_cache_max_size == 3



from __future__ import annotations

from typing import Any
from typing import Mapping
from typing import cast

from jsonschema.exceptions import best_match
from jsonschema.protocols import Validator
from referencing import Registry

from openapi_schema_validator._caches import ValidatorCache
from openapi_schema_validator._dialects import OAS31_BASE_DIALECT_ID
from openapi_schema_validator._dialects import OAS32_BASE_DIALECT_ID
from openapi_schema_validator.validators import OAS32Validator
from openapi_schema_validator.validators import (
    build_enforce_properties_required_validator,
)
from openapi_schema_validator.validators import check_openapi_schema

_LOCAL_ONLY_REGISTRY = Registry()
_VALIDATOR_CACHE = ValidatorCache()


def _check_schema(
    cls: type[Validator],
    schema: dict[str, Any],
) -> None:
    meta_schema = getattr(cls, "META_SCHEMA", None)
    # jsonschema's default check_schema path does not accept a custom
    # registry, so for OAS dialects we use the package registry
    # explicitly to keep metaschema resolution local and deterministic.
    if isinstance(meta_schema, dict) and meta_schema.get("$id") in (
        OAS31_BASE_DIALECT_ID,
        OAS32_BASE_DIALECT_ID,
    ):
        check_openapi_schema(cls, schema)
    else:
        cls.check_schema(schema)


def validate(
    instance: Any,
    schema: Mapping[str, Any],
    cls: type[Validator] = OAS32Validator,
    *args: Any,
    allow_remote_references: bool = False,
    check_schema: bool = True,
    enforce_properties_required: bool = False,
    **kwargs: Any,
) -> None:
    """
    Validate an instance against a given schema using the specified
    validator class.

    Unlike direct ``Validator(schema).validate(instance)`` usage, this helper
    checks schema validity first.
    Invalid schemas therefore raise ``SchemaError`` before any instance
    validation occurs.

    Args:
        instance: Value to validate against ``schema``.
        schema: OpenAPI schema mapping used for validation. Local references
            (``#/...``) are resolved against this mapping.
        cls: Validator class to use. Defaults to ``OAS32Validator``.
        *args: Positional arguments forwarded to ``cls`` constructor.
        allow_remote_references: If ``True`` and no explicit ``registry`` is
            provided, allow jsonschema's default remote reference retrieval
            behavior.
        check_schema: If ``True`` (default), validate the provided schema
            before validating ``instance``. If ``False``, skip schema
            validation and run instance validation directly.
        enforce_properties_required: If ``True``, all properties declared in
            the schema's ``properties`` object are strictly required to be
            present in the instance (except those marked as ``writeOnly`` or
            ``readOnly`` where appropriate), regardless of the schema's
            ``required`` array. Defaults to ``False``.
        **kwargs: Keyword arguments forwarded to ``cls`` constructor
            (for example ``registry`` and ``format_checker``). If omitted,
            a local-only empty ``Registry`` is used to avoid implicit remote
            reference retrieval.

    Raises:
        jsonschema.exceptions.SchemaError: If ``schema`` is invalid.
        jsonschema.exceptions.ValidationError: If ``instance`` is invalid.
    """
    if enforce_properties_required:
        cls = build_enforce_properties_required_validator(cls)  # type: ignore[arg-type]

    schema_dict = cast(dict[str, Any], schema)

    validator_kwargs = kwargs.copy()
    if not allow_remote_references:
        validator_kwargs.setdefault("registry", _LOCAL_ONLY_REGISTRY)

    key = _VALIDATOR_CACHE.build_key(
        schema=schema_dict,
        cls=cls,
        args=args,
        kwargs=validator_kwargs,
        allow_remote_references=allow_remote_references,
    )

    cached = _VALIDATOR_CACHE.get(key)

    if cached is None:
        if check_schema:
            _check_schema(cls, schema_dict)

        validator = cls(schema_dict, *args, **validator_kwargs)
        cached = _VALIDATOR_CACHE.set(
            key,
            validator=validator,
            schema_checked=check_schema,
        )
    elif check_schema and not cached.schema_checked:
        _check_schema(cls, schema_dict)
        _VALIDATOR_CACHE.mark_schema_checked(key)
    else:
        _VALIDATOR_CACHE.touch(key)

    error = best_match(
        cached.validator.evolve(schema=schema_dict).iter_errors(instance)
    )
    if error is not None:
        raise error


def clear_validate_cache() -> None:
    _VALIDATOR_CACHE.clear()



from dataclasses import dataclass
from typing import Any

from referencing import Registry
from referencing import Resource
from referencing.jsonschema import DRAFT202012

from openapi_schema_validator import OAS30Validator
from openapi_schema_validator import OAS31Validator
from openapi_schema_validator import OAS32Validator
from openapi_schema_validator import oas30_format_checker
from openapi_schema_validator import oas31_format_checker
from openapi_schema_validator import oas32_format_checker


@dataclass(frozen=True)
class BenchmarkCase:
    name: str
    validator_class: Any
    schema: dict[str, Any]
    instance: Any
    validator_kwargs: dict[str, Any]


def build_cases() -> list[BenchmarkCase]:
    name_schema = Resource.from_contents(
        {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "type": "string",
        }
    )
    age_schema = DRAFT202012.create_resource(
        {
            "type": "integer",
            "format": "int32",
            "minimum": 0,
            "maximum": 120,
        }
    )
    registry = Registry().with_resources(
        [
            ("urn:name-schema", name_schema),
            ("urn:age-schema", age_schema),
        ]
    )

    return [
        BenchmarkCase(
            name="oas32_simple_object",
            validator_class=OAS32Validator,
            schema={
                "type": "object",
                "required": ["name"],
                "properties": {
                    "name": {"type": "string"},
                    "enabled": {"type": "boolean"},
                },
                "additionalProperties": False,
            },
            instance={"name": "svc", "enabled": True},
            validator_kwargs={"format_checker": oas32_format_checker},
        ),
        BenchmarkCase(
            name="oas31_prefix_items",
            validator_class=OAS31Validator,
            schema={
                "type": "array",
                "prefixItems": [
                    {"type": "number"},
                    {"type": "string"},
                    {"enum": ["Street", "Avenue", "Boulevard"]},
                    {"enum": ["NW", "NE", "SW", "SE"]},
                ],
                "items": False,
            },
            instance=[1600, "Pennsylvania", "Avenue", "NW"],
            validator_kwargs={"format_checker": oas31_format_checker},
        ),
        BenchmarkCase(
            name="oas30_nullable",
            validator_class=OAS30Validator,
            schema={"type": "string", "nullable": True},
            instance=None,
            validator_kwargs={"format_checker": oas30_format_checker},
        ),
        BenchmarkCase(
            name="oas30_discriminator",
            validator_class=OAS30Validator,
            schema={
                "$ref": "#/components/schemas/Route",
                "components": {
                    "schemas": {
                        "MountainHiking": {
                            "type": "object",
                            "properties": {
                                "discipline": {
                                    "type": "string",
                                    "enum": [
                                        "mountain_hiking",
                                        "MountainHiking",
                                    ],
                                },
                                "length": {"type": "integer"},
                            },
                            "required": ["discipline", "length"],
                        },
                        "AlpineClimbing": {
                            "type": "object",
                            "properties": {
                                "discipline": {
                                    "type": "string",
                                    "enum": ["alpine_climbing"],
                                },
                                "height": {"type": "integer"},
                            },
                            "required": ["discipline", "height"],
                        },
                        "Route": {
                            "oneOf": [
                                {
                                    "$ref": (
                                        "#/components/schemas/"
                                        "MountainHiking"
                                    )
                                },
                                {
                                    "$ref": (
                                        "#/components/schemas/"
                                        "AlpineClimbing"
                                    )
                                },
                            ],
                            "discriminator": {
                                "propertyName": "discipline",
                                "mapping": {
                                    "mountain_hiking": (
                                        "#/components/schemas/"
                                        "MountainHiking"
                                    ),
                                    "alpine_climbing": (
                                        "#/components/schemas/"
                                        "AlpineClimbing"
                                    ),
                                },
                            },
                        },
                    }
                },
            },
            instance={"discipline": "mountain_hiking", "length": 10},
            validator_kwargs={"format_checker": oas30_format_checker},
        ),
        BenchmarkCase(
            name="oas32_registry_refs",
            validator_class=OAS32Validator,
            schema={
                "type": "object",
                "required": ["name"],
                "properties": {
                    "name": {"$ref": "urn:name-schema"},
                    "age": {"$ref": "urn:age-schema"},
                },
                "additionalProperties": False,
            },
            instance={"name": "John", "age": 23},
            validator_kwargs={
                "format_checker": oas32_format_checker,
                "registry": registry,
            },
        ),
    ]



from openapi_schema_validator._dialects import OAS31_BASE_DIALECT_ID
from openapi_schema_validator._dialects import OAS32_BASE_DIALECT_ID
from openapi_schema_validator._format import oas30_format_checker
from openapi_schema_validator._format import oas30_strict_format_checker
from openapi_schema_validator._format import oas31_format_checker
from openapi_schema_validator._format import oas32_format_checker
from openapi_schema_validator.shortcuts import validate
from openapi_schema_validator.validators import OAS30ReadValidator
from openapi_schema_validator.validators import OAS30StrictValidator
from openapi_schema_validator.validators import OAS30Validator
from openapi_schema_validator.validators import OAS30WriteValidator
from openapi_schema_validator.validators import OAS31Validator
from openapi_schema_validator.validators import OAS32Validator

__author__ = "Artur Maciag"
__email__ = "maciag.artur@gmail.com"
__version__ = "0.9.0"
__url__ = "https://github.com/python-openapi/openapi-schema-validator"
__license__ = "3-clause BSD License"

__all__ = [
    "validate",
    "OAS30ReadValidator",
    "OAS30StrictValidator",
    "OAS30WriteValidator",
    "OAS30Validator",
    "oas30_format_checker",
    "oas30_strict_format_checker",
    "OAS31Validator",
    "oas31_format_checker",
    "OAS32Validator",
    "oas32_format_checker",
    "OAS31_BASE_DIALECT_ID",
    "OAS32_BASE_DIALECT_ID",
]



import binascii
from base64 import b64decode
from base64 import b64encode
from numbers import Number

from jsonschema._format import FormatChecker

from openapi_schema_validator._regex import is_valid_regex


def is_int32(instance: object) -> bool:
    # bool inherits from int, so ensure bools aren't reported as ints
    if isinstance(instance, bool):
        return True
    if not isinstance(instance, int):
        return True
    return ~(1 << 31) < instance < 1 << 31


def is_int64(instance: object) -> bool:
    # bool inherits from int, so ensure bools aren't reported as ints
    if isinstance(instance, bool):
        return True
    if not isinstance(instance, int):
        return True
    return ~(1 << 63) < instance < 1 << 63


def is_float(instance: object) -> bool:
    # bool inherits from int
    if isinstance(instance, int):
        return True
    if not isinstance(instance, Number):
        return True
    return isinstance(instance, float)


def is_double(instance: object) -> bool:
    # bool inherits from int
    if isinstance(instance, int):
        return True
    if not isinstance(instance, Number):
        return True
    # float has double precision in Python
    # It's double in CPython and Jython
    return isinstance(instance, float)


def is_binary_strict(instance: object) -> bool:
    # Strict: only accepts base64-encoded strings, not raw bytes
    if isinstance(instance, bytes):
        return False
    if isinstance(instance, str):
        try:
            b64decode(instance)
            return True
        except Exception:
            return False
    return True


def is_binary_pragmatic(instance: object) -> bool:
    # Pragmatic: accepts bytes (common in Python) or base64-encoded strings
    if isinstance(instance, (str, bytes)):
        return True
    return True


def is_byte(instance: object) -> bool:
    if not isinstance(instance, (str, bytes)):
        return True
    if isinstance(instance, str):
        instance = instance.encode("ascii", errors="strict")

    try:
        b64decode(instance, validate=True)
    except (binascii.Error, ValueError):
        return False
    return True


def is_password(instance: object) -> bool:
    # A hint to UIs to obscure input
    return True


def is_regex(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    return is_valid_regex(instance)


oas30_format_checker = FormatChecker()
oas30_format_checker.checks("int32")(is_int32)
oas30_format_checker.checks("int64")(is_int64)
oas30_format_checker.checks("float")(is_float)
oas30_format_checker.checks("double")(is_double)
oas30_format_checker.checks("binary")(is_binary_pragmatic)
oas30_format_checker.checks("byte", (binascii.Error, TypeError))(is_byte)
oas30_format_checker.checks("password")(is_password)
oas30_format_checker.checks("regex")(is_regex)

oas30_strict_format_checker = FormatChecker()
oas30_strict_format_checker.checks("int32")(is_int32)
oas30_strict_format_checker.checks("int64")(is_int64)
oas30_strict_format_checker.checks("float")(is_float)
oas30_strict_format_checker.checks("double")(is_double)
oas30_strict_format_checker.checks("binary")(is_binary_strict)
oas30_strict_format_checker.checks("byte", (binascii.Error, TypeError))(
    is_byte
)
oas30_strict_format_checker.checks("password")(is_password)
oas30_strict_format_checker.checks("regex")(is_regex)

oas31_format_checker = FormatChecker()
oas31_format_checker.checks("int32")(is_int32)
oas31_format_checker.checks("int64")(is_int64)
oas31_format_checker.checks("float")(is_float)
oas31_format_checker.checks("double")(is_double)
oas31_format_checker.checks("password")(is_password)
oas31_format_checker.checks("regex")(is_regex)

# OAS 3.2 uses the same format checks as OAS 3.1
oas32_format_checker = FormatChecker()
oas32_format_checker.checks("int32")(is_int32)
oas32_format_checker.checks("int64")(is_int64)
oas32_format_checker.checks("float")(is_float)
oas32_format_checker.checks("double")(is_double)
oas32_format_checker.checks("password")(is_password)
oas32_format_checker.checks("regex")(is_regex)



from functools import lru_cache
from typing import Any
from typing import Iterator
from typing import Mapping
from typing import cast

from jsonschema import _keywords
from jsonschema import _legacy_keywords
from jsonschema.exceptions import SchemaError
from jsonschema.exceptions import ValidationError
from jsonschema.protocols import Validator
from jsonschema.validators import Draft202012Validator
from jsonschema.validators import create
from jsonschema.validators import extend
from jsonschema.validators import validator_for

from openapi_schema_validator import _format as oas_format
from openapi_schema_validator import _keywords as oas_keywords
from openapi_schema_validator import _types as oas_types
from openapi_schema_validator._dialects import OAS31_BASE_DIALECT_ID
from openapi_schema_validator._dialects import OAS31_BASE_DIALECT_METASCHEMA
from openapi_schema_validator._dialects import OAS32_BASE_DIALECT_ID
from openapi_schema_validator._dialects import OAS32_BASE_DIALECT_METASCHEMA
from openapi_schema_validator._dialects import register_openapi_dialect
from openapi_schema_validator._specifications import (
    REGISTRY as OPENAPI_SPECIFICATIONS,
)
from openapi_schema_validator._types import oas31_type_checker

_CHECK_SCHEMA_UNSET = object()


def check_openapi_schema(
    cls: Any,
    schema: Any,
    format_checker: Any = _CHECK_SCHEMA_UNSET,
) -> None:
    if format_checker is _CHECK_SCHEMA_UNSET:
        format_checker = cls.FORMAT_CHECKER

    validator_class = validator_for(cls.META_SCHEMA, default=cls)

    validator_for_metaschema = validator_class(
        cls.META_SCHEMA,
        format_checker=format_checker,
        registry=OPENAPI_SPECIFICATIONS,
    )

    for error in validator_for_metaschema.iter_errors(schema):
        raise SchemaError.create_from(error)


def _oas30_id_of(schema: Any) -> str:
    if isinstance(schema, dict):
        return schema.get("id", "")  # type: ignore[no-any-return]
    return ""


OAS30_VALIDATORS = cast(
    Any,
    {
        "multipleOf": _keywords.multipleOf,
        # exclusiveMaximum supported inside maximum_draft3_draft4
        "maximum": _legacy_keywords.maximum_draft3_draft4,
        # exclusiveMinimum supported inside minimum_draft3_draft4
        "minimum": _legacy_keywords.minimum_draft3_draft4,
        "maxLength": _keywords.maxLength,
        "minLength": _keywords.minLength,
        "pattern": oas_keywords.pattern,
        "maxItems": _keywords.maxItems,
        "minItems": _keywords.minItems,
        "uniqueItems": _keywords.uniqueItems,
        "maxProperties": _keywords.maxProperties,
        "minProperties": _keywords.minProperties,
        "enum": _keywords.enum,
        # adjusted to OAS
        "type": oas_keywords.type,
        "allOf": oas_keywords.allOf,
        "oneOf": oas_keywords.oneOf,
        "anyOf": oas_keywords.anyOf,
        "not": _keywords.not_,
        "items": oas_keywords.items,
        "properties": _keywords.properties,
        "required": oas_keywords.required,
        "additionalProperties": oas_keywords.additionalProperties,
        # TODO: adjust description
        "format": oas_keywords.format,
        # TODO: adjust default
        "$ref": _keywords.ref,
        # fixed OAS fields
        "discriminator": oas_keywords.not_implemented,
        "readOnly": oas_keywords.not_implemented,
        "writeOnly": oas_keywords.not_implemented,
        "xml": oas_keywords.not_implemented,
        "externalDocs": oas_keywords.not_implemented,
        "example": oas_keywords.not_implemented,
        "deprecated": oas_keywords.not_implemented,
    },
)


def _build_oas30_validator() -> Any:
    return create(
        meta_schema=OPENAPI_SPECIFICATIONS.contents(
            "http://json-schema.org/draft-04/schema#",
        ),
        validators=OAS30_VALIDATORS,
        type_checker=oas_types.oas30_type_checker,
        format_checker=oas_format.oas30_format_checker,
        # NOTE: version causes conflict with global jsonschema validator
        # See https://github.com/python-openapi/openapi-schema-validator/pull/12
        # version="oas30",
        id_of=_oas30_id_of,
    )


def _build_oas31_validator() -> Any:
    validator = extend(
        Draft202012Validator,
        {
            # adjusted to OAS
            "pattern": oas_keywords.pattern,
            "description": oas_keywords.not_implemented,
            # fixed OAS fields
            # discriminator is annotation-only in OAS 3.1+
            "discriminator": oas_keywords.not_implemented,
            "xml": oas_keywords.not_implemented,
            "externalDocs": oas_keywords.not_implemented,
            "example": oas_keywords.not_implemented,
        },
        type_checker=oas31_type_checker,
        format_checker=oas_format.oas31_format_checker,
    )
    return register_openapi_dialect(
        validator=validator,
        dialect_id=OAS31_BASE_DIALECT_ID,
        version_name="oas31",
        metaschema=OAS31_BASE_DIALECT_METASCHEMA,
    )


def _build_oas32_validator() -> Any:
    validator = extend(
        OAS31Validator,
        {},
        format_checker=oas_format.oas32_format_checker,
    )
    return register_openapi_dialect(
        validator=validator,
        dialect_id=OAS32_BASE_DIALECT_ID,
        version_name="oas32",
        metaschema=OAS32_BASE_DIALECT_METASCHEMA,
    )


OAS30Validator = _build_oas30_validator()
OAS30StrictValidator = extend(
    OAS30Validator,
    validators={
        "type": oas_keywords.strict_type,
    },
    type_checker=oas_types.oas30_type_checker,
    format_checker=oas_format.oas30_strict_format_checker,
    # NOTE: version causes conflict with global jsonschema validator
    # See https://github.com/python-openapi/openapi-schema-validator/pull/12
    # version="oas30-strict",
)
OAS30ReadValidator = extend(
    OAS30Validator,
    validators={
        "required": oas_keywords.read_required,
        "writeOnly": oas_keywords.read_writeOnly,
    },
)
OAS30WriteValidator = extend(
    OAS30Validator,
    validators={
        "required": oas_keywords.write_required,
        "readOnly": oas_keywords.write_readOnly,
    },
)

OAS31Validator = _build_oas31_validator()
OAS32Validator = _build_oas32_validator()

# These validator classes are generated via jsonschema create/extend, so there
# is no simpler hook to inject registry-aware schema checking while preserving
# each class's FORMAT_CHECKER. Override check_schema on each class to keep
# OpenAPI metaschema resolution local and to apply optional ecma-regex
# behavior consistently across OAS 3.0/3.1/3.2.
OAS30Validator.check_schema = classmethod(check_openapi_schema)
OAS31Validator.check_schema = classmethod(check_openapi_schema)
OAS32Validator.check_schema = classmethod(check_openapi_schema)


@lru_cache(maxsize=None)
def build_enforce_properties_required_validator(
    validator_class: Any,
) -> type[Validator]:
    properties_validator = validator_class.VALIDATORS.get("properties")
    required_validator = validator_class.VALIDATORS.get("required")

    def enforce_properties(
        validator: Any,
        properties: Any,
        instance: Any,
        schema: Mapping[str, Any],
    ) -> Iterator[Any]:
        if properties_validator is not None:
            yield from properties_validator(
                validator, properties, instance, schema
            )

        if not validator.is_type(instance, "object"):
            return

        if required_validator is not None:
            schema_required = (
                schema.get("required", []) if isinstance(schema, dict) else []
            )
            missing_props = [
                p for p in properties.keys() if p not in schema_required
            ]
            if missing_props:
                yield from required_validator(
                    validator, missing_props, instance, schema
                )

    extended_validator = extend(
        validator_class,
        validators={"properties": enforce_properties},
    )
    if hasattr(validator_class, "check_schema"):
        extended_validator.check_schema = classmethod(
            validator_class.check_schema.__func__
        )
    return cast(type[Validator], extended_validator)



import re
import warnings
from base64 import b64encode
from typing import Any
from typing import cast
from unittest.mock import patch

import pytest
from jsonschema import SchemaError
from jsonschema import ValidationError
from jsonschema.exceptions import (
    _WrappedReferencingError as WrappedReferencingError,
)
from jsonschema.validators import Draft202012Validator
from jsonschema.validators import extend
from jsonschema.validators import validator_for
from referencing import Registry
from referencing import Resource
from referencing.exceptions import InvalidAnchor
from referencing.exceptions import NoSuchAnchor
from referencing.exceptions import PointerToNowhere
from referencing.jsonschema import DRAFT202012

from openapi_schema_validator import OAS30ReadValidator
from openapi_schema_validator import OAS30StrictValidator
from openapi_schema_validator import OAS30Validator
from openapi_schema_validator import OAS30WriteValidator
from openapi_schema_validator import OAS31Validator
from openapi_schema_validator import OAS32Validator
from openapi_schema_validator import oas30_format_checker
from openapi_schema_validator import oas30_strict_format_checker
from openapi_schema_validator import oas31_format_checker
from openapi_schema_validator import oas32_format_checker
from openapi_schema_validator import validate
from openapi_schema_validator._dialects import OAS31_BASE_DIALECT_ID
from openapi_schema_validator._dialects import OAS31_BASE_DIALECT_METASCHEMA
from openapi_schema_validator._dialects import OAS32_BASE_DIALECT_ID
from openapi_schema_validator._dialects import OAS32_BASE_DIALECT_METASCHEMA
from openapi_schema_validator._dialects import register_openapi_dialect
from openapi_schema_validator._regex import has_ecma_regex


class TestOAS30ValidatorFormatChecker:
    @pytest.fixture
    def format_checker(self):
        return OAS30Validator.FORMAT_CHECKER

    def test_required_checkers(self, format_checker):
        required_formats_set = {
            "int32",
            "int64",
            "float",
            "double",
            "byte",
            "binary",
            "date",
            "date-time",
            "password",
        }
        assert required_formats_set.issubset(
            set(format_checker.checkers.keys())
        )


class BaseTestOASValidatorValidate:
    @pytest.mark.parametrize(
        "format,value",
        [
            ("int32", "test"),
            ("int32", True),
            ("int32", 3.12),
            ("int32", ["test"]),
            ("int64", "test"),
            ("int64", True),
            ("int64", 3.12),
            ("int64", ["test"]),
            ("float", "test"),
            ("float", 3),
            ("float", True),
            ("float", ["test"]),
            ("double", "test"),
            ("double", 3),
            ("double", True),
            ("double", ["test"]),
            ("password", 3.12),
            ("password", True),
            ("password", 3),
            ("password", ["test"]),
        ],
    )
    def test_formats_ignored(
        self, format, value, validator_class, format_checker
    ):
        schema = {"format": format}
        validator = validator_class(schema, format_checker=format_checker)

        result = validator.validate(value)

        assert result is None

    @pytest.mark.parametrize("format", ["float", "double"])
    @pytest.mark.parametrize("value", [3, 3.14, 1.0])
    def test_number_float_and_double_valid(
        self, format, value, validator_class, format_checker
    ):
        schema = {"type": "number", "format": format}
        validator = validator_class(schema, format_checker=format_checker)

        result = validator.validate(value)

        assert result is None

    @pytest.mark.parametrize("value", ["test"])
    def test_string(self, validator_class, value):
        schema = {"type": "string"}
        validator = validator_class(schema)

        result = validator.validate(value)

        assert result is None

    @pytest.mark.parametrize("value", [True, 3, 3.12, None])
    def test_string_invalid(self, validator_class, value):
        schema = {"type": "string"}
        validator = validator_class(schema)

        with pytest.raises(ValidationError):
            validator.validate(value)

    def test_invalid_pattern_raises_expected_error(self, validator_class):
        schema = {"type": "string", "pattern": "["}
        validator = validator_class(schema)

        if has_ecma_regex():
            with pytest.raises(
                ValidationError,
                match="is not a valid regular expression",
            ):
                validator.validate("foo")
        else:
            with pytest.raises(re.error):
                validator.validate("foo")

    def test_invalid_pattern_rejected_by_validate_helper(
        self, validator_class
    ):
        schema = {"type": "string", "pattern": "["}

        with pytest.raises(SchemaError, match="is not a 'regex'"):
            validate("foo", schema, cls=validator_class)

    @pytest.mark.skipif(
        not has_ecma_regex(), reason="requires optional ecma-regex extra"
    )
    def test_z_escape_behaves_as_ecma_literal_escape(self, validator_class):
        schema = {"type": "string", "pattern": r"^foo\z"}
        validator = validator_class(schema)

        with pytest.raises(ValidationError, match="does not match"):
            validator.validate("foo")

        result = validator.validate("fooz")

        assert result is None

        result = validate("fooz", schema, cls=validator_class)

        assert result is None

    @pytest.mark.skipif(
        not has_ecma_regex(), reason="requires optional ecma-regex extra"
    )
    def test_escaped_z_pattern_is_allowed_with_ecma_regex(
        self, validator_class
    ):
        schema = {"type": "string", "pattern": r"^foo\\z$"}
        validator = validator_class(schema)

        result = validator.validate(r"foo\z")

        assert result is None

    def test_referencing(self, validator_class):
        name_schema = Resource.from_contents(
            {
                "$schema": "https://json-schema.org/draft/2020-12/schema",
                "type": "string",
            }
        )
        age_schema = DRAFT202012.create_resource(
            {
                "type": "integer",
                "format": "int32",
                "minimum": 0,
                "maximum": 120,
            }
        )
        birth_date_schema = Resource.from_contents(
            {
                "type": "string",
                "format": "date",
            },
            default_specification=DRAFT202012,
        )
        registry = Registry().with_resources(
            [
                ("urn:name-schema", name_schema),
                ("urn:age-schema", age_schema),
                ("urn:birth-date-schema", birth_date_schema),
            ],
        )
        schema = {
            "type": "object",
            "required": ["name"],
            "properties": {
                "name": {"$ref": "urn:name-schema"},
                "age": {"$ref": "urn:age-schema"},
                "birth-date": {"$ref": "urn:birth-date-schema"},
            },
            "additionalProperties": False,
        }

        validator = validator_class(schema, registry=registry)
        result = validator.validate({"name": "John", "age": 23})

        assert result is None


class TestOAS30ValidatorValidate(BaseTestOASValidatorValidate):
    @pytest.fixture
    def validator_class(self):
        return OAS30Validator

    @pytest.fixture
    def format_checker(self):
        return oas30_format_checker

    @pytest.mark.parametrize(
        "format,value",
        [
            ("binary", True),
            ("binary", 3),
            ("binary", 3.12),
            ("binary", ["test"]),
            ("byte", True),
            ("byte", 3),
            ("byte", 3.12),
            ("byte", ["test"]),
        ],
    )
    def test_oas30_formats_ignored(
        self, format, value, validator_class, format_checker
    ):
        schema = {"format": format}
        validator = validator_class(schema, format_checker=format_checker)

        result = validator.validate(value)

        assert result is None

    @pytest.mark.parametrize("value", [b"test"])
    def test_string_disallow_binary(self, validator_class, value):
        schema = {"type": "string"}
        validator = validator_class(schema)

        with pytest.raises(ValidationError):
            validator.validate(value)

    @pytest.mark.parametrize("value", [b"test"])
    def test_string_binary_valid(self, validator_class, format_checker, value):
        schema = {"type": "string", "format": "binary"}
        validator = validator_class(schema, format_checker=format_checker)

        result = validator.validate(value)

        assert result is None

    @pytest.mark.parametrize("value", [True, 3, 3.12, None])
    def test_string_binary_invalid(
        self, validator_class, format_checker, value
    ):
        schema = {"type": "string", "format": "binary"}
        validator = validator_class(schema, format_checker=format_checker)

        with pytest.raises(ValidationError):
            validator.validate(value)

    @pytest.mark.parametrize(
        "schema_type",
        [
            "boolean",
            "array",
            "integer",
            "number",
            "string",
        ],
    )
    def test_null(self, validator_class, schema_type):
        schema = {"type": schema_type}
        validator = validator_class(schema)
        value = None

        with pytest.raises(ValidationError):
            validator.validate(value)

    @pytest.mark.parametrize("is_nullable", [True, False])
    def test_nullable_untyped(self, validator_class, is_nullable):
        schema = {"nullable": is_nullable}
        validator = validator_class(schema)
        value = None

        result = validator.validate(value)

        assert result is None

    @pytest.mark.parametrize(
        "schema_type",
        [
            "boolean",
            "array",
            "integer",
            "number",
            "string",
        ],
    )
    def test_nullable(self, validator_class, schema_type):
        schema = {"type": schema_type, "nullable": True}
        validator = validator_class(schema)
        value = None

        result = validator.validate(value)

        assert result is None

    def test_nullable_enum_without_none(self, validator_class):
        schema = {"type": "integer", "nullable": True, "enum": [1, 2, 3]}
        validator = validator_class(schema)
        value = None

        with pytest.raises(ValidationError):
            validator.validate(value)

    def test_nullable_enum_with_none(self, validator_class):
        schema = {"type": "integer", "nullable": True, "enum": [1, 2, 3, None]}
        validator = validator_class(schema)
        value = None

        result = validator.validate(value)

        assert result is None

    @pytest.mark.parametrize(
        "value",
        [
            b64encode(b"string").decode(),
            b64encode(b"\x00\x01\x02").decode(),
            "",
            "AQ==",
            "SGVsbG8=",
        ],
    )
    def test_string_format_byte_valid(self, validator_class, value):
        schema = {"type": "string", "format": "byte"}
        validator = validator_class(
            schema, format_checker=oas30_format_checker
        )

        result = validator.validate(value)

        assert result is None

    @pytest.mark.parametrize(
        "value",
        [
            "string",
            "SGVsbG8",
            "SGVsbG8===",
            "SGVsbG8$",
            "SGVsbG8 ",
            "SGVsbG8\n",
            "SGVsbG8_",
            "SGVsbG8-",
        ],
    )
    def test_string_format_byte_invalid(self, validator_class, value):
        schema = {"type": "string", "format": "byte"}
        validator = validator_class(
            schema, format_checker=oas30_format_checker
        )

        with pytest.raises(ValidationError, match="is not a 'byte'"):
            validator.validate(value)

    def test_allof_required(self, validator_class):
        schema = {
            "allOf": [
                {
                    "type": "object",
                    "properties": {"some_prop": {"type": "string"}},
                },
                {"type": "object", "required": ["some_prop"]},
            ]
        }
        validator = validator_class(
            schema, format_checker=oas30_format_checker
        )
        with pytest.raises(
            ValidationError, match="'some_prop' is a required property"
        ):
            validator.validate({"another_prop": "bla"})

    def test_required(self, validator_class):
        schema = {
            "type": "object",
            "properties": {"some_prop": {"type": "string"}},
            "required": ["some_prop"],
        }

        validator = validator_class(
            schema, format_checker=oas30_format_checker
        )
        with pytest.raises(
            ValidationError, match="'some_prop' is a required property"
        ):
            validator.validate({"another_prop": "bla"})
        assert validator.validate({"some_prop": "hello"}) is None

    def test_oneof_required(self, validator_class):
        instance = {
            "n3IwfId": "string",
        }
        schema = {
            "type": "object",
            "properties": {
                "n3IwfId": {"type": "string"},
                "wagfId": {"type": "string"},
            },
            "oneOf": [
                {"required": ["n3IwfId"]},
                {"required": ["wagfId"]},
            ],
        }
        validator = validator_class(
            schema, format_checker=oas30_format_checker
        )
        result = validator.validate(instance)
        assert result is None

    @pytest.mark.parametrize(
        "mapping_ref",
        [
            "#/components/schemas/Missing",
            "#missing-anchor",
            "#bad/frag",
        ],
    )
    def test_discriminator_handles_unresolvable_reference_kinds(
        self, mapping_ref
    ):
        schema = {
            "oneOf": [{"$ref": "#/components/schemas/MountainHiking"}],
            "discriminator": {
                "propertyName": "discipline",
                "mapping": {"mountain_hiking": mapping_ref},
            },
            "components": {
                "schemas": {
                    "MountainHiking": {
                        "type": "object",
                        "properties": {
                            "discipline": {"type": "string"},
                            "length": {"type": "integer"},
                        },
                        "required": ["discipline", "length"],
                    },
                },
            },
        }

        validator = OAS30Validator(
            schema,
            format_checker=oas30_format_checker,
        )
        with pytest.raises(
            ValidationError,
            match=f"reference '{mapping_ref}' could not be resolved",
        ):
            validator.validate(
                {
                    "discipline": "mountain_hiking",
                    "length": 10,
                }
            )

    @pytest.mark.parametrize(
        "mapping_ref, expected_cause",
        [
            ("#/components/schemas/Missing", PointerToNowhere),
            ("#missing-anchor", NoSuchAnchor),
            ("#bad/frag", InvalidAnchor),
        ],
    )
    def test_discriminator_unresolvable_reference_causes(
        self, mapping_ref, expected_cause
    ):
        schema = {
            "oneOf": [{"$ref": "#/components/schemas/MountainHiking"}],
            "discriminator": {
                "propertyName": "discipline",
                "mapping": {"mountain_hiking": mapping_ref},
            },
            "components": {
                "schemas": {
                    "MountainHiking": {
                        "type": "object",
                        "properties": {
                            "discipline": {"type": "string"},
                            "length": {"type": "integer"},
                        },
                        "required": ["discipline", "length"],
                    },
                },
            },
        }

        validator = OAS30Validator(
            schema,
            format_checker=oas30_format_checker,
        )

        with pytest.raises(WrappedReferencingError) as exc_info:
            cast(Any, validator)._validate_reference(
                ref=mapping_ref,
                instance={"discipline": "mountain_hiking", "length": 10},
            )

        assert isinstance(exc_info.value.__cause__, expected_cause)

    @pytest.mark.parametrize(
        "schema_type",
        [
            "oneOf",
            "anyOf",
            "allOf",
        ],
    )
    def test_oneof_discriminator(self, validator_class, schema_type):
        # We define a few components schemas
        components = {
            "MountainHiking": {
                "type": "object",
                "properties": {
                    "discipline": {
                        "type": "string",
                        # we allow both the explicitly matched mountain_hiking discipline
                        # and the implicitly matched MoutainHiking discipline
                        "enum": ["mountain_hiking", "MountainHiking"],
                    },
                    "length": {
                        "type": "integer",
                    },
                },
                "required": ["discipline", "length"],
            },
            "AlpineClimbing": {
                "type": "object",
                "properties": {
                    "discipline": {
                        "type": "string",
                        "enum": ["alpine_climbing"],
                    },
                    "height": {
                        "type": "integer",
                    },
                },
                "required": ["discipline", "height"],
            },
            "Route": {
                # defined later
            },
        }
        components["Route"][schema_type] = [
            {"$ref": "#/components/schemas/MountainHiking"},
            {"$ref": "#/components/schemas/AlpineClimbing"},
        ]

        # Add the components in a minimalis schema
        schema = {
            "$ref": "#/components/schemas/Route",
            "components": {"schemas": components},
        }

        if schema_type != "allOf":
            # use jsonschema validator when no discriminator is defined
            validator = validator_class(
                schema, format_checker=oas30_format_checker
            )
            with pytest.raises(
                ValidationError,
                match="is not valid under any of the given schemas",
            ):
                validator.validate(
                    {"something": "matching_none_of_the_schemas"}
                )
                assert False

        if schema_type == "anyOf":
            # use jsonschema validator when no discriminator is defined
            validator = validator_class(
                schema, format_checker=oas30_format_checker
            )
            with pytest.raises(
                ValidationError,
                match="is not valid under any of the given schemas",
            ):
                validator.validate(
                    {"something": "matching_none_of_the_schemas"}
                )
                assert False

        discriminator = {
            "propertyName": "discipline",
            "mapping": {
                "mountain_hiking": "#/components/schemas/MountainHiking",
                "alpine_climbing": "#/components/schemas/AlpineClimbing",
            },
        }
        schema["components"]["schemas"]["Route"][
            "discriminator"
        ] = discriminator

        # Optional: check we return useful result when the schema is wrong
        validator = validator_class(
            schema, format_checker=oas30_format_checker
        )

        with pytest.raises(
            ValidationError,
            match="is not of type 'object'",
        ):
            validator.validate("not-an-object")

        with pytest.raises(
            ValidationError, match="does not contain discriminating property"
        ):
            validator.validate({"something": "missing"})
            assert False

        # Check we get a non-generic, somehow usable, error message when a discriminated schema is failing
        with pytest.raises(
            ValidationError, match="'bad_string' is not of type 'integer'"
        ):
            validator.validate(
                {"discipline": "mountain_hiking", "length": "bad_string"}
            )
            assert False

        # Check explicit MountainHiking resolution
        validator.validate({"discipline": "mountain_hiking", "length": 10})

        # Check implicit MountainHiking resolution
        validator.validate({"discipline": "MountainHiking", "length": 10})

        # Check non resolvable implicit schema
        with pytest.raises(
            ValidationError,
            match="reference '#/components/schemas/other' could not be resolved",
        ):
            result = validator.validate({"discipline": "other"})
            assert False

    @pytest.mark.parametrize("is_nullable", [True, False])
    def test_nullable_ref(self, validator_class, is_nullable):
        """
        Tests that a field that points to a schema reference is null checked based on the $ref schema rather than
        on this schema
        :param is_nullable:  if the schema is marked as nullable. If not, validate an exception is raised on None
        """
        schema = {
            "$ref": "#/$defs/Pet",
            "$defs": {
                "NullableText": {"type": "string", "nullable": is_nullable},
                "Pet": {
                    "properties": {
                        "testfield": {"$ref": "#/$defs/NullableText"},
                    },
                },
            },
        }
        validator = validator_class(
            schema,
            format_checker=oas30_format_checker,
        )

        result = validator.validate({"testfield": "John"})
        assert result is None

        if is_nullable:
            result = validator.validate({"testfield": None})
            assert result is None
        else:
            with pytest.raises(
                ValidationError,
                match="None for not nullable",
            ):
                validator.validate({"testfield": None})
                assert False

    @pytest.mark.parametrize(
        "schema_type, not_nullable_regex",
        [
            ("oneOf", "None is not valid under any of the given schemas"),
            ("anyOf", "None is not valid under any of the given schemas"),
            ("allOf", "None for not nullable"),
        ],
    )
    @pytest.mark.parametrize("is_nullable", [True, False])
    def test_nullable_schema_combos(
        self, validator_class, is_nullable, schema_type, not_nullable_regex
    ):
        """
        This test ensures that nullablilty semantics are correct for oneOf, anyOf and allOf
        Specifically, nullable should checked on the children schemas
        :param is_nullable:  if the schema is marked as nullable. If not, validate an exception is raised on None
        :param schema_type: the schema type to validate
        :param not_nullable_regex: the expected raised exception if fields are marked as not nullable
        """
        schema = {
            "$ref": "#/$defs/Pet",
            "$defs": {
                "NullableText": {
                    "type": "string",
                    "nullable": (
                        False if schema_type == "oneOf" else is_nullable
                    ),
                },
                "NullableEnum": {
                    "type": "string",
                    "nullable": is_nullable,
                    "enum": ["John", "Alice", None],
                },
                "Pet": {
                    "properties": {
                        "testfield": {
                            schema_type: [
                                {"$ref": "#/$defs/NullableText"},
                                {"$ref": "#/$defs/NullableEnum"},
                            ]
                        }
                    },
                },
            },
        }
        validator = validator_class(
            schema,
            format_checker=oas30_format_checker,
        )

        if is_nullable:
            result = validator.validate({"testfield": None})
            assert result is None
        else:
            with pytest.raises(ValidationError, match=not_nullable_regex):
                validator.validate({"testfield": None})
                assert False


class TestOAS30ReadWriteValidatorValidate:
    def test_read_only(self):
        schema = {
            "type": "object",
            "properties": {"some_prop": {"type": "string", "readOnly": True}},
        }

        validator = OAS30WriteValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        with pytest.raises(
            ValidationError,
            match="Tried to write read-only property with hello",
        ):
            validator.validate({"some_prop": "hello"})
        validator = OAS30ReadValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        assert validator.validate({"some_prop": "hello"}) is None
        validator = OAS30Validator(
            schema,
            format_checker=oas30_format_checker,
        )
        assert validator.validate({"some_prop": "hello"}) is None

    def test_write_only(self):
        schema = {
            "type": "object",
            "properties": {"some_prop": {"type": "string", "writeOnly": True}},
        }

        validator = OAS30ReadValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        with pytest.raises(
            ValidationError,
            match="Tried to read write-only property with hello",
        ):
            validator.validate({"some_prop": "hello"})
        validator = OAS30WriteValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        assert validator.validate({"some_prop": "hello"}) is None
        validator = OAS30Validator(
            schema,
            format_checker=oas30_format_checker,
        )
        assert validator.validate({"some_prop": "hello"}) is None

    def test_required_read_only(self):
        schema = {
            "type": "object",
            "properties": {"some_prop": {"type": "string", "readOnly": True}},
            "required": ["some_prop"],
        }

        validator = OAS30ReadValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        with pytest.raises(
            ValidationError, match="'some_prop' is a required property"
        ):
            validator.validate({"another_prop": "hello"})
        validator = OAS30WriteValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        assert validator.validate({"another_prop": "hello"}) is None

    def test_required_write_only(self):
        schema = {
            "type": "object",
            "properties": {"some_prop": {"type": "string", "writeOnly": True}},
            "required": ["some_prop"],
        }

        validator = OAS30WriteValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        with pytest.raises(
            ValidationError, match="'some_prop' is a required property"
        ):
            validator.validate({"another_prop": "hello"})
        validator = OAS30ReadValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        assert validator.validate({"another_prop": "hello"}) is None

    def test_read_only_false(self):
        schema = {
            "type": "object",
            "properties": {"some_prop": {"type": "string", "readOnly": False}},
        }

        validator = OAS30WriteValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        assert validator.validate({"some_prop": "hello"}) is None

    def test_write_only_false(self):
        schema = {
            "type": "object",
            "properties": {
                "some_prop": {"type": "string", "writeOnly": False}
            },
        }

        validator = OAS30ReadValidator(
            schema,
            format_checker=oas30_format_checker,
        )
        assert validator.validate({"some_prop": "hello"}) is None


class TestOAS31ValidatorFormatChecker:
    @pytest.fixture
    def format_checker(self):
        return OAS31Validator.FORMAT_CHECKER

    def test_required_checkers(self, format_checker):
        required_formats_set = {
            # standard formats
            "int32",
            "int64",
            "float",
            "double",
            "password",
        }
        assert required_formats_set.issubset(
            set(format_checker.checkers.keys())
        )


class TestOAS31ValidatorValidate(BaseTestOASValidatorValidate):
    @pytest.fixture
    def validator_class(self):
        return OAS31Validator

    @pytest.fixture
    def format_checker(self):
        return oas31_format_checker

    @pytest.mark.parametrize("value", [b"test"])
    def test_string_disallow_binary(self, validator_class, value):
        schema = {"type": "string"}
        validator = validator_class(schema)

        with pytest.raises(ValidationError):
            validator.validate(value)

    @pytest.mark.parametrize("value", [b"test"])
    def test_string_binary_rejects_bytes(
        self, validator_class, format_checker, value
    ):
        schema = {"type": "string", "format": "binary"}
        validator = validator_class(schema, format_checker=format_checker)

        with pytest.raises(ValidationError):
            validator.validate(value)

    @pytest.mark.parametrize("value", [True, 3, 3.12, None])
    def test_string_binary_invalid(
        self, validator_class, format_checker, value
    ):
        schema = {"type": "string", "format": "binary"}
        validator = validator_class(schema, format_checker=format_checker)

        with pytest.raises(ValidationError):
            validator.validate(value)

    @pytest.mark.parametrize(
        "schema_type",
        [
            "boolean",
            "array",
            "integer",
            "number",
            "string",
        ],
    )
    def test_null(self, validator_class, schema_type):
        schema = {"type": schema_type}
        validator = validator_class(schema)
        value = None

        with pytest.raises(ValidationError):
            validator.validate(value)

    @pytest.mark.parametrize(
        "schema_type",
        [
            "boolean",
            "array",
            "integer",
            "number",
            "string",
        ],
    )
    def test_nullable(self, validator_class, schema_type):
        schema = {"type": [schema_type, "null"]}
        validator = validator_class(schema)
        value = None

        result = validator.validate(value)

        assert result is None

    def test_schema_validation(self, validator_class, format_checker):
        schema = {
            "type": "object",
            "required": ["name"],
            "properties": {
                "name": {"type": "string"},
                "age": {
                    "type": "integer",
                    "format": "int32",
                    "minimum": 0,
                    "nullable": True,
                },
                "birth-date": {
                    "type": "string",
                    "format": "date",
                },
            },
            "additionalProperties": False,
        }
        validator = validator_class(
            schema,
            format_checker=format_checker,
        )

        result = validator.validate({"name": "John", "age": 23})
        assert result is None

        with pytest.raises(ValidationError) as excinfo:
            validator.validate({"name": "John", "city": "London"})

        error = "Additional properties are not allowed ('city' was unexpected)"
        assert error in str(excinfo.value)

        with pytest.raises(ValidationError) as excinfo:
            validator.validate({"name": "John", "birth-date": "-12"})

        error = "'-12' is not a 'date'"
        assert error in str(excinfo.value)

    def test_schema_ref(self, validator_class, format_checker):
        schema = {
            "$ref": "#/$defs/Pet",
            "$defs": {
                "Pet": {
                    "required": ["id", "name"],
                    "properties": {
                        "id": {"type": "integer", "format": "int64"},
                        "name": {"type": "string"},
                        "tag": {"type": "string"},
                    },
                }
            },
        }
        validator = validator_class(
            schema,
            format_checker=format_checker,
        )

        result = validator.validate({"id": 1, "name": "John"})
        assert result is None

        with pytest.raises(ValidationError) as excinfo:
            validator.validate({"name": "John"})

        error = "'id' is a required property"
        assert error in str(excinfo.value)

    @pytest.mark.parametrize(
        "value",
        [
            [1600, "Pennsylvania", "Avenue", "NW"],
            [1600, "Pennsylvania", "Avenue"],
        ],
    )
    def test_array_prefixitems(self, validator_class, format_checker, value):
        schema = {
            "type": "array",
            "prefixItems": [
                {"type": "number"},
                {"type": "string"},
                {"enum": ["Street", "Avenue", "Boulevard"]},
                {"enum": ["NW", "NE", "SW", "SE"]},
            ],
            "items": False,
        }
        validator = validator_class(
            schema,
            format_checker=format_checker,
        )

        result = validator.validate(value)

        assert result is None

    @pytest.mark.parametrize(
        "value",
        [
            [1600, "Pennsylvania", "Avenue", "NW", "Washington"],
        ],
    )
    def test_array_prefixitems_invalid(self, validator_class, value):
        schema = {
            "type": "array",
            "prefixItems": [
                {"type": "number"},
                {"type": "string"},
                {"enum": ["Street", "Avenue", "Boulevard"]},
                {"enum": ["NW", "NE", "SW", "SE"]},
            ],
            "items": False,
        }
        validator = validator_class(
            schema,
            format_checker=oas31_format_checker,
        )

        with pytest.raises(ValidationError) as excinfo:
            validator.validate(value)

        errors = [
            # jsonschema < 4.20.0
            "Expected at most 4 items, but found 5",
            # jsonschema >= 4.20.0
            "Expected at most 4 items but found 1 extra",
        ]
        assert any(error in str(excinfo.value) for error in errors)

    def test_discriminator_is_annotation_only(self, validator_class):
        schema = {
            "components": {
                "schemas": {
                    "A": {
                        "type": "object",
                        "properties": {"kind": {"type": "string"}},
                        "required": ["kind"],
                    },
                    "B": {
                        "type": "object",
                        "properties": {
                            "kind": {"type": "string"},
                            "other": {"type": "string"},
                        },
                        "required": ["kind"],
                    },
                }
            },
            "oneOf": [
                {"$ref": "#/components/schemas/A"},
                {"$ref": "#/components/schemas/B"},
            ],
            "discriminator": {"propertyName": "kind"},
        }

        validator = validator_class(schema)

        # A payload valid for both schemas A and B
        instance = {"kind": "B"}

        # oneOf fails because it matches BOTH A and B, discriminator does not restrict it
        with pytest.raises(ValidationError):
            validator.validate(instance)

    @pytest.mark.parametrize(
        "mapping_ref",
        [
            "#/components/schemas/Missing",
            "#missing-anchor",
            "#bad/frag",
        ],
    )
    def test_discriminator_unresolvable_reference_ignored(
        self, validator_class, mapping_ref
    ):
        schema = {
            "oneOf": [{"$ref": "#/components/schemas/MountainHiking"}],
            "discriminator": {
                "propertyName": "discipline",
                "mapping": {"mountain_hiking": mapping_ref},
            },
            "components": {
                "schemas": {
                    "MountainHiking": {
                        "type": "object",
                        "properties": {
                            "discipline": {"type": "string"},
                            "length": {"type": "integer"},
                        },
                        "required": ["discipline", "length"],
                    },
                },
            },
        }

        validator = validator_class(schema)

        # It should not raise any referencing errors because discriminator mapping is annotation-only
        validator.validate(
            {"discipline": "mountain_hiking", "length": 10},
        )


class TestOAS32ValidatorValidate(TestOAS31ValidatorValidate):
    """OAS 3.2 uses the OAS 3.2 published dialect resources."""

    @pytest.fixture
    def validator_class(self):
        return OAS32Validator

    @pytest.fixture
    def format_checker(self):
        return oas32_format_checker

    def test_validator_is_distinct_from_oas31(self):
        assert OAS32Validator is not OAS31Validator

    def test_format_checker_is_distinct_from_oas31(self):
        assert oas32_format_checker is not oas31_format_checker

    def test_validator_shares_oas31_behavior(self):
        assert OAS32Validator.VALIDATORS == OAS31Validator.VALIDATORS

    def test_validator_has_oas32_dialect_metaschema(self):
        assert OAS32Validator.META_SCHEMA["$id"] == OAS32_BASE_DIALECT_ID

    def test_format_validation_int32(self, validator_class):
        schema = {"type": "integer", "format": "int32"}
        validator = validator_class(
            schema, format_checker=oas32_format_checker
        )

        result = validator.validate(42)
        assert result is None

        with pytest.raises(ValidationError):
            validator.validate(9999999999)

    def test_format_validation_date(self, validator_class):
        schema = {"type": "string", "format": "date"}
        validator = validator_class(
            schema, format_checker=oas32_format_checker
        )

        result = validator.validate("2024-01-15")
        assert result is None

        with pytest.raises(ValidationError):
            validator.validate("not-a-date")

    def test_schema_with_allof(self, validator_class):
        schema = {
            "allOf": [
                {"type": "object", "properties": {"id": {"type": "integer"}}},
                {"type": "object", "properties": {"name": {"type": "string"}}},
            ]
        }
        validator = validator_class(schema)

        result = validator.validate({"id": 1, "name": "test"})
        assert result is None

        with pytest.raises(ValidationError):
            validator.validate({"id": "not-an-integer"})

    def test_check_schema_accepts_oas32_discriminator_default_mapping(self):
        schema = {
            "type": "object",
            "discriminator": {
                "propertyName": "kind",
                "defaultMapping": "#/components/schemas/Pet",
            },
        }

        OAS32Validator.check_schema(schema)

    def test_oas31_check_schema_rejects_discriminator_default_mapping(self):
        schema = {
            "type": "object",
            "discriminator": {
                "propertyName": "kind",
                "defaultMapping": "#/components/schemas/Pet",
            },
        }

        with pytest.raises(SchemaError):
            OAS31Validator.check_schema(schema)

    def test_oas32_check_schema_does_not_fetch_remote_metaschemas(self):
        schema = {
            "type": "object",
            "discriminator": {
                "propertyName": "kind",
                "defaultMapping": "#/components/schemas/Pet",
            },
        }

        with patch("urllib.request.urlopen") as urlopen:
            OAS32Validator.check_schema(schema)

        urlopen.assert_not_called()


class TestOAS30StrictValidator:
    """
    Tests for OAS30StrictValidator which follows OAS spec strictly:
    - type: string only accepts str (not bytes)
    - format: binary also only accepts str (no special bytes handling)
    """

    def test_strict_string_rejects_bytes(self):
        """Strict validator rejects bytes for plain string type."""
        schema = {"type": "string"}
        validator = OAS30StrictValidator(schema)

        with pytest.raises(ValidationError):
            validator.validate(b"test")

    def test_strict_string_accepts_str(self):
        """Strict validator accepts str for string type."""
        schema = {"type": "string"}
        validator = OAS30StrictValidator(schema)

        result = validator.validate("test")
        assert result is None

    def test_strict_binary_format_rejects_bytes(self):
        """Strict validator rejects bytes even with binary format."""
        schema = {"type": "string", "format": "binary"}
        validator = OAS30StrictValidator(
            schema, format_checker=oas30_format_checker
        )

        with pytest.raises(ValidationError):
            validator.validate(b"test")

    def test_strict_binary_format_rejects_str(self):
        """
        Strict validator with binary format rejects strings.
        Binary format is for bytes in OAS, not plain strings.
        """
        schema = {"type": "string", "format": "binary"}
        validator = OAS30StrictValidator(
            schema, format_checker=oas30_strict_format_checker
        )

        # Binary format expects actual binary data (bytes in Python)
        # Plain strings fail format validation because they are not valid base64
        # Note: "test" is actually valid base64, so use "not base64" which is not
        with pytest.raises(ValidationError, match="is not a 'binary'"):
            validator.validate("not base64")


class TestValidatorForDiscovery:
    def test_oas31_base_dialect_resolves_to_oas31_validator(self):
        schema = {"$schema": OAS31_BASE_DIALECT_ID}

        validator_class = validator_for(schema)

        assert validator_class is OAS31Validator

    def test_oas31_base_dialect_discovery_has_no_deprecation_warning(self):
        schema = {"$schema": OAS31_BASE_DIALECT_ID}

        with warnings.catch_warnings(record=True) as caught:
            warnings.simplefilter("always")
            validator_for(schema)

        assert not any(
            issubclass(warning.category, DeprecationWarning)
            for warning in caught
        )

    def test_oas32_base_dialect_resolves_to_oas32_validator(self):
        schema = {"$schema": OAS32_BASE_DIALECT_ID}

        validator_class = validator_for(schema)

        assert validator_class is OAS32Validator

    def test_oas32_base_dialect_discovery_has_no_deprecation_warning(self):
        schema = {"$schema": OAS32_BASE_DIALECT_ID}

        with warnings.catch_warnings(record=True) as caught:
            warnings.simplefilter("always")
            validator_for(schema)

        assert not any(
            issubclass(warning.category, DeprecationWarning)
            for warning in caught
        )

    def test_oas31_base_dialect_keeps_oas_keyword_behavior(self):
        schema = {
            "$schema": OAS31_BASE_DIALECT_ID,
            "type": "object",
            "required": ["kind"],
            "properties": {"kind": {"type": "string"}},
            "discriminator": {"propertyName": "kind"},
            "xml": {"name": "Pet"},
            "example": {"kind": "cat"},
        }

        validator_class = validator_for(schema)
        validator = validator_class(
            schema, format_checker=oas31_format_checker
        )

        result = validator.validate({"kind": "cat"})

        assert result is None

    def test_draft_2020_12_discovery_is_unchanged(self):
        schema = {"$schema": "https://json-schema.org/draft/2020-12/schema"}

        validator_class = validator_for(schema)

        assert validator_class is Draft202012Validator

    def test_openapi_dialect_registration_is_idempotent(self):
        register_openapi_dialect(
            validator=OAS31Validator,
            dialect_id=OAS31_BASE_DIALECT_ID,
            version_name="oas31",
            metaschema=OAS31_BASE_DIALECT_METASCHEMA,
        )
        register_openapi_dialect(
            validator=OAS31Validator,
            dialect_id=OAS31_BASE_DIALECT_ID,
            version_name="oas31",
            metaschema=OAS31_BASE_DIALECT_METASCHEMA,
        )

        validator_class = validator_for({"$schema": OAS31_BASE_DIALECT_ID})

        assert validator_class is OAS31Validator

    def test_openapi_dialect_registration_does_not_replace_validator(self):
        another_oas31_validator = extend(OAS31Validator, {})

        registered_validator = register_openapi_dialect(
            validator=another_oas31_validator,
            dialect_id=OAS31_BASE_DIALECT_ID,
            version_name="oas31",
            metaschema=OAS31_BASE_DIALECT_METASCHEMA,
        )

        assert registered_validator is OAS31Validator
        assert (
            validator_for({"$schema": OAS31_BASE_DIALECT_ID}) is OAS31Validator
        )

    def test_openapi_oas32_dialect_registration_is_idempotent(self):
        register_openapi_dialect(
            validator=OAS32Validator,
            dialect_id=OAS32_BASE_DIALECT_ID,
            version_name="oas32",
            metaschema=OAS32_BASE_DIALECT_METASCHEMA,
        )
        register_openapi_dialect(
            validator=OAS32Validator,
            dialect_id=OAS32_BASE_DIALECT_ID,
            version_name="oas32",
            metaschema=OAS32_BASE_DIALECT_METASCHEMA,
        )

        validator_class = validator_for({"$schema": OAS32_BASE_DIALECT_ID})

        assert validator_class is OAS32Validator



from __future__ import annotations

import argparse
import gc
import json
import platform
import statistics
import time
import tracemalloc
from datetime import datetime
from datetime import timezone
from pathlib import Path
from typing import Any

from benchmarks.cases import BenchmarkCase
from benchmarks.cases import build_cases
from openapi_schema_validator.shortcuts import clear_validate_cache
from openapi_schema_validator.shortcuts import validate


def _measure_compile_time_ms(
    case: BenchmarkCase,
    rounds: int,
) -> float:
    samples: list[float] = []
    for _ in range(rounds):
        start_ns = time.perf_counter_ns()
        case.validator_class(
            case.schema,
            **case.validator_kwargs,
        )
        elapsed_ms = (time.perf_counter_ns() - start_ns) / 1_000_000
        samples.append(elapsed_ms)
    return statistics.median(samples)


def _measure_first_validate_ms(case: BenchmarkCase) -> float:
    validator = case.validator_class(case.schema, **case.validator_kwargs)
    start_ns = time.perf_counter_ns()
    validator.validate(case.instance)
    return (time.perf_counter_ns() - start_ns) / 1_000_000


def _measure_compiled_validate_per_second(
    case: BenchmarkCase,
    iterations: int,
    warmup: int,
) -> float:
    validator = case.validator_class(case.schema, **case.validator_kwargs)
    for _ in range(warmup):
        validator.validate(case.instance)

    start_ns = time.perf_counter_ns()
    for _ in range(iterations):
        validator.validate(case.instance)
    elapsed = (time.perf_counter_ns() - start_ns) / 1_000_000_000
    return iterations / elapsed


def _measure_helper_validate_per_second(
    case: BenchmarkCase,
    iterations: int,
    warmup: int,
    *,
    check_schema: bool,
) -> float:
    clear_validate_cache()
    for _ in range(warmup):
        validate(
            case.instance,
            case.schema,
            cls=case.validator_class,
            check_schema=check_schema,
            **case.validator_kwargs,
        )

    start_ns = time.perf_counter_ns()
    for _ in range(iterations):
        validate(
            case.instance,
            case.schema,
            cls=case.validator_class,
            check_schema=check_schema,
            **case.validator_kwargs,
        )
    elapsed = (time.perf_counter_ns() - start_ns) / 1_000_000_000
    return iterations / elapsed


def _measure_peak_memory_kib(
    case: BenchmarkCase,
    iterations: int,
) -> float:
    validator = case.validator_class(case.schema, **case.validator_kwargs)
    tracemalloc.start()
    for _ in range(iterations):
        validator.validate(case.instance)
    _, peak = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    return peak / 1024


def _measure_case(
    case: BenchmarkCase,
    iterations: int,
    warmup: int,
    compile_rounds: int,
) -> dict[str, Any]:
    gc_enabled = gc.isenabled()
    gc.disable()

    try:
        return {
            "name": case.name,
            "validator_class": case.validator_class.__name__,
            "compile_ms": _measure_compile_time_ms(case, compile_rounds),
            "first_validate_ms": _measure_first_validate_ms(case),
            "compiled_validations_per_second": (
                _measure_compiled_validate_per_second(
                    case,
                    iterations,
                    warmup,
                )
            ),
            "helper_validations_per_second": (
                _measure_helper_validate_per_second(
                    case,
                    iterations,
                    warmup,
                    check_schema=True,
                )
            ),
            "helper_trusted_validations_per_second": (
                _measure_helper_validate_per_second(
                    case,
                    iterations,
                    warmup,
                    check_schema=False,
                )
            ),
            "compiled_peak_memory_kib": _measure_peak_memory_kib(
                case,
                max(iterations, 100),
            ),
        }
    finally:
        if gc_enabled:
            gc.enable()


def _build_report(
    cases: list[BenchmarkCase],
    iterations: int,
    warmup: int,
    compile_rounds: int,
) -> dict[str, Any]:
    results = [
        _measure_case(
            case,
            iterations=iterations,
            warmup=warmup,
            compile_rounds=compile_rounds,
        )
        for case in cases
    ]
    return {
        "timestamp_utc": datetime.now(timezone.utc).isoformat(),
        "python_version": platform.python_version(),
        "platform": platform.platform(),
        "benchmark_parameters": {
            "iterations": iterations,
            "warmup": warmup,
            "compile_rounds": compile_rounds,
        },
        "cases": results,
    }


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Run validation performance benchmarks.",
    )
    parser.add_argument(
        "--iterations",
        type=int,
        default=1000,
        help="Measured validation iterations per case.",
    )
    parser.add_argument(
        "--warmup",
        type=int,
        default=100,
        help="Warmup iterations per case.",
    )
    parser.add_argument(
        "--compile-rounds",
        type=int,
        default=50,
        help="Schema compile measurements per case.",
    )
    parser.add_argument(
        "--output",
        type=Path,
        default=Path("reports/benchmarks/python-baseline.json"),
        help="Path to write JSON benchmark report.",
    )
    return parser.parse_args()


def main() -> int:
    args = _parse_args()
    cases = build_cases()
    report = _build_report(
        cases,
        iterations=args.iterations,
        warmup=args.warmup,
        compile_rounds=args.compile_rounds,
    )

    output_path = args.output
    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(json.dumps(report, indent=2), encoding="utf-8")

    print(f"Saved benchmark report to {output_path}")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import inspect
import re
from unittest.mock import patch

import pytest
from jsonschema.exceptions import SchemaError
from jsonschema.exceptions import ValidationError
from referencing import Registry
from referencing import Resource

from openapi_schema_validator import OAS32Validator
from openapi_schema_validator import validate
from openapi_schema_validator._regex import has_ecma_regex
from openapi_schema_validator.settings import reset_settings_cache
from openapi_schema_validator.shortcuts import clear_validate_cache
from openapi_schema_validator.validators import OAS30ReadValidator
from openapi_schema_validator.validators import OAS30Validator
from openapi_schema_validator.validators import OAS30WriteValidator
from openapi_schema_validator.validators import OAS31Validator
from openapi_schema_validator.validators import OAS32Validator


@pytest.fixture(scope="function")
def schema():
    return {
        "type": "object",
        "properties": {
            "email": {"type": "string"},
            "enabled": {
                "type": "boolean",
            },
        },
        "example": {"enabled": False, "email": "foo@bar.com"},
    }


@pytest.fixture(autouse=True)
def clear_validate_cache_fixture():
    reset_settings_cache()
    clear_validate_cache()
    yield
    clear_validate_cache()
    reset_settings_cache()


def test_validate_does_not_add_nullable_to_schema(schema):
    """
    Verify that calling validate does not add the 'nullable' key to the schema
    """
    validate({"email": "foo@bar.com"}, schema)
    assert "nullable" not in schema["properties"]["email"].keys()


def test_validate_does_not_mutate_schema(schema):
    """
    Verify that calling validate does not mutate the schema
    """
    original_schema = schema.copy()
    validate({"email": "foo@bar.com"}, schema)
    assert schema == original_schema


def test_validate_does_not_fetch_remote_metaschemas(schema):
    with patch("urllib.request.urlopen") as urlopen:
        validate({"email": "foo@bar.com"}, schema)

    urlopen.assert_not_called()


def test_validate_defaults_to_oas32_validator():
    signature = inspect.signature(validate)

    assert signature.parameters["cls"].default is OAS32Validator


def test_oas32_validate_does_not_fetch_remote_metaschemas(schema):
    with patch("urllib.request.urlopen") as urlopen:
        validate({"email": "foo@bar.com"}, schema, cls=OAS32Validator)

    urlopen.assert_not_called()


def test_validate_blocks_implicit_remote_http_references_by_default():
    schema = {"$ref": "http://example.com/remote-schema.json"}

    with patch("urllib.request.urlopen") as urlopen:
        with pytest.raises(Exception, match="Unresolvable"):
            validate({}, schema)

    urlopen.assert_not_called()


def test_validate_blocks_implicit_file_references_by_default():
    schema = {"$ref": "file:///etc/hosts"}

    with patch("urllib.request.urlopen") as urlopen:
        with pytest.raises(Exception, match="Unresolvable"):
            validate({}, schema)

    urlopen.assert_not_called()


def test_validate_local_references_still_work_by_default():
    schema = {"$defs": {"Value": {"type": "integer"}}, "$ref": "#/$defs/Value"}

    with patch("urllib.request.urlopen") as urlopen:
        result = validate(1, schema)

    assert result is None
    urlopen.assert_not_called()


def test_validate_honors_explicit_registry():
    schema = {
        "type": "object",
        "properties": {"name": {"$ref": "urn:name-schema"}},
    }
    name_schema = Resource.from_contents(
        {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "type": "string",
        }
    )
    registry = Registry().with_resources(
        [("urn:name-schema", name_schema)],
    )

    result = validate({"name": "John"}, schema, registry=registry)

    assert result is None


def test_validate_can_allow_implicit_remote_references():
    schema = {"$ref": "http://example.com/remote-schema.json"}

    with patch("urllib.request.urlopen") as urlopen:
        with pytest.raises(Exception):
            validate({}, schema, allow_remote_references=True)

    assert urlopen.called


def test_validate_skip_schema_check():
    schema = {"type": "string", "pattern": "["}

    with pytest.raises(SchemaError, match="is not a 'regex'"):
        validate("foo", schema)

    if has_ecma_regex():
        with pytest.raises(
            ValidationError, match="is not a valid regular expression"
        ):
            validate("foo", schema, check_schema=False)
    else:
        with pytest.raises(re.error):
            validate("foo", schema, check_schema=False)


def test_validate_cache_avoids_rechecking_schema(schema):
    with patch(
        "openapi_schema_validator.shortcuts.check_openapi_schema"
    ) as check_schema_mock:
        validate({"email": "foo@bar.com"}, schema, cls=OAS32Validator)
        validate({"email": "foo@bar.com"}, schema, cls=OAS32Validator)

    check_schema_mock.assert_called_once()


def test_validate_cache_promotes_unchecked_validator(schema):
    with patch(
        "openapi_schema_validator.shortcuts.check_openapi_schema"
    ) as check_schema_mock:
        validate(
            {"email": "foo@bar.com"},
            schema,
            cls=OAS32Validator,
            check_schema=False,
        )
        validate({"email": "foo@bar.com"}, schema, cls=OAS32Validator)
        validate({"email": "foo@bar.com"}, schema, cls=OAS32Validator)

    check_schema_mock.assert_called_once()


def test_validate_cache_max_size_from_env(monkeypatch):
    schema_a = {"type": "string"}
    schema_b = {"type": "integer"}

    monkeypatch.setenv(
        "OPENAPI_SCHEMA_VALIDATOR_COMPILED_VALIDATOR_CACHE_MAX_SIZE",
        "1",
    )
    reset_settings_cache()

    with patch(
        "openapi_schema_validator.shortcuts.check_openapi_schema"
    ) as check_schema_mock:
        validate("foo", schema_a, cls=OAS32Validator)
        validate(1, schema_b, cls=OAS32Validator)
        validate("foo", schema_a, cls=OAS32Validator)

    assert check_schema_mock.call_count == 3


@pytest.mark.parametrize(
    "schema, cls, instance, enforce, expected_error",
    [
        (
            {
                "type": "object",
                "properties": {
                    "id": {"type": "string"},
                    "nickname": {"type": "string"},
                },
                "required": ["id"],
            },
            OAS30Validator,
            {"id": "42"},
            False,
            None,
        ),
        (
            {
                "type": "object",
                "properties": {
                    "id": {"type": "string"},
                    "nickname": {"type": "string"},
                },
                "required": ["id"],
            },
            OAS30Validator,
            {"id": "42"},
            True,
            "'nickname' is a required property",
        ),
        (
            {
                "type": "object",
                "properties": {
                    "id": {"type": "string", "readOnly": True},
                    "password": {"type": "string", "writeOnly": True},
                    "normal": {"type": "string"},
                },
            },
            OAS30ReadValidator,
            {"id": "123"},
            True,
            "'normal' is a required property",
        ),
        (
            {
                "type": "object",
                "properties": {
                    "id": {"type": "string", "readOnly": True},
                    "password": {"type": "string", "writeOnly": True},
                    "normal": {"type": "string"},
                },
            },
            OAS30ReadValidator,
            {"normal": "abc"},
            True,
            "'id' is a required property",
        ),
        (
            {
                "type": "object",
                "properties": {
                    "id": {"type": "string", "readOnly": True},
                    "password": {"type": "string", "writeOnly": True},
                    "normal": {"type": "string"},
                },
            },
            OAS30ReadValidator,
            {"id": "123", "normal": "abc"},
            True,
            None,
        ),
        (
            {
                "type": "object",
                "properties": {
                    "id": {"type": "string", "readOnly": True},
                    "password": {"type": "string", "writeOnly": True},
                    "normal": {"type": "string"},
                },
            },
            OAS30WriteValidator,
            {"normal": "abc"},
            True,
            "'password' is a required property",
        ),
        (
            {
                "type": "object",
                "properties": {
                    "id": {"type": "string", "readOnly": True},
                    "password": {"type": "string", "writeOnly": True},
                    "normal": {"type": "string"},
                },
            },
            OAS30WriteValidator,
            {"password": "secret"},
            True,
            "'normal' is a required property",
        ),
        (
            {
                "type": "object",
                "properties": {
                    "id": {"type": "string", "readOnly": True},
                    "password": {"type": "string", "writeOnly": True},
                    "normal": {"type": "string"},
                },
            },
            OAS30WriteValidator,
            {"password": "secret", "normal": "abc"},
            True,
            None,
        ),
        (
            {
                "type": "object",
                "properties": {
                    "foo": True,
                },
            },
            OAS31Validator,
            {},
            False,
            None,
        ),
        (
            {
                "type": "object",
                "properties": {
                    "foo": True,
                },
            },
            OAS31Validator,
            {},
            True,
            "'foo' is a required property",
        ),
        (
            {
                "type": "object",
                "properties": {
                    "foo": {"type": "string"},
                },
            },
            OAS32Validator,
            {},
            False,
            None,
        ),
        (
            {
                "type": "object",
                "properties": {
                    "foo": {"type": "string"},
                },
            },
            OAS32Validator,
            {},
            True,
            "'foo' is a required property",
        ),
    ],
)
def test_enforce_properties_required(
    schema, cls, instance, enforce, expected_error
):
    if expected_error:
        with pytest.raises(ValidationError) as exc:
            validate(
                instance,
                schema,
                cls=cls,
                enforce_properties_required=enforce,
            )
        assert expected_error in str(exc.value)
    else:
        validate(
            instance, schema, cls=cls, enforce_properties_required=enforce
        )



import json
from importlib.resources import files
from typing import Any
from typing import Iterator

from jsonschema_specifications import REGISTRY as JSONSCHEMA_REGISTRY
from referencing import Resource

__all__ = ["REGISTRY"]


def _iter_schema_files() -> Iterator[Any]:
    schema_root = files(__package__).joinpath("schemas")
    stack = [schema_root]

    while stack:
        current = stack.pop()
        for child in current.iterdir():
            if child.name.startswith("."):
                continue
            if child.is_dir():
                stack.append(child)
                continue
            yield child


def _load_schemas() -> Iterator[Resource]:
    for path in _iter_schema_files():
        contents = json.loads(path.read_text(encoding="utf-8"))
        yield Resource.from_contents(contents)


#: A `referencing.Registry` containing all official jsonschema resources
#: plus openapi resources.
REGISTRY = (_load_schemas() @ JSONSCHEMA_REGISTRY).crawl()



from collections import OrderedDict
from dataclasses import dataclass
from threading import RLock
from typing import Any
from typing import Hashable
from typing import Mapping

from jsonschema.protocols import Validator

from openapi_schema_validator.settings import get_settings


@dataclass
class CachedValidator:
    validator: Any
    schema_checked: bool


class ValidatorCache:
    def __init__(self) -> None:
        self._cache: OrderedDict[Hashable, CachedValidator] = OrderedDict()
        self._lock = RLock()

    def _freeze_value(self, value: Any) -> Hashable:
        if isinstance(value, dict):
            return tuple(
                sorted(
                    (str(key), self._freeze_value(item))
                    for key, item in value.items()
                )
            )
        if isinstance(value, list):
            return tuple(self._freeze_value(item) for item in value)
        if isinstance(value, tuple):
            return tuple(self._freeze_value(item) for item in value)
        if isinstance(value, set):
            return tuple(
                sorted(
                    (self._freeze_value(item) for item in value),
                    key=repr,
                )
            )
        if isinstance(value, (str, bytes, int, float, bool, type(None))):
            return value
        return ("id", id(value))

    def _schema_fingerprint(self, schema: Mapping[str, Any]) -> Hashable:
        return self._freeze_value(dict(schema))

    def build_key(
        self,
        schema: Mapping[str, Any],
        cls: type[Validator],
        args: tuple[Any, ...],
        kwargs: Mapping[str, Any],
        allow_remote_references: bool,
    ) -> Hashable:
        return (
            cls,
            allow_remote_references,
            self._schema_fingerprint(schema),
            self._freeze_value(args),
            self._freeze_value(dict(kwargs)),
        )

    def get(self, key: Hashable) -> CachedValidator | None:
        with self._lock:
            return self._cache.get(key)

    def set(
        self,
        key: Hashable,
        *,
        validator: Any,
        schema_checked: bool,
    ) -> CachedValidator:
        cached = CachedValidator(
            validator=validator,
            schema_checked=schema_checked,
        )
        with self._lock:
            self._cache[key] = cached
            self._cache.move_to_end(key)
            self._prune_if_needed()
        return cached

    def mark_schema_checked(self, key: Hashable) -> None:
        with self._lock:
            cached = self._cache.get(key)
            if cached is None:
                return
            cached.schema_checked = True
            self._cache.move_to_end(key)

    def touch(self, key: Hashable) -> None:
        with self._lock:
            if key in self._cache:
                self._cache.move_to_end(key)

    def clear(self) -> None:
        with self._lock:
            self._cache.clear()

    def _prune_if_needed(self) -> None:
        max_size = get_settings().compiled_validator_cache_max_size
        while len(self._cache) > max_size:
            self._cache.popitem(last=False)
